"""Internal utilities shared across funcache modules."""

import inspect
import os
from typing import Any, Callable, Dict, Tuple


def normalize_args(
    func: Callable, args: Tuple[Any, ...], kwargs: Dict[str, Any]
) -> Dict[str, Any]:
    """Convert positional arguments to keyword arguments based on function signature.

    This merges positional args into a unified kwargs dict for consistent
    cache key lookup by parameter name.
    """
    merged = dict(kwargs)
    params = list(inspect.signature(func).parameters.items())
    for i, (name, param) in enumerate(params):
        if name not in merged:
            merged[name] = args[i] if i < len(args) else param.default
    return merged


def ensure_gitignore(directory: str) -> None:
    """Create a .gitignore with '*' in the given directory if one doesn't exist."""
    os.makedirs(directory, exist_ok=True)
    ignore_file = os.path.join(directory, ".gitignore")
    if not os.path.exists(ignore_file):
        with open(ignore_file, "w") as f:
            f.write("*")



from typing import Callable

from cachebox import FIFOCache, LFUCache, LRUCache, RRCache, TTLCache, VTTLCache, cached

__all__ = [
    "cache",
    "lru_cache",
    "ttl_cache",
    "vttl_cache",
    "lfu_cache",
    "fifo_cache",
    "rr_cache",
]


def cache(func: Callable, /) -> Callable:
    """Simple LRU cache decorator with a default maxsize of 1000."""
    return cached(LRUCache(maxsize=1000))(func)


def vttl_cache(maxsize: int = 1000, ttl: int = 60) -> Callable:
    """VTTLCache: removes expired elements lazily when accessed.

    VTTLCache: 在访问时才惰性移除已过期的缓存元素。
    """
    return lambda func: cached(VTTLCache(maxsize=maxsize, ttl=ttl))(func)


def ttl_cache(maxsize: int = 1000, ttl: int = 60) -> Callable:
    """TTLCache: automatically removes expired elements.

    TTLCache：自动移除已过期的缓存元素。
    """
    return lambda func: cached(TTLCache(maxsize=maxsize, ttl=ttl))(func)


def lru_cache(maxsize: int = 1000) -> Callable:
    """LRUCache: removes the least recently used element.

    LRUCache：移除缓存中自上次访问以来时间最长的元素。
    """
    return lambda func: cached(LRUCache(maxsize=maxsize))(func)


def lfu_cache(maxsize: int = 1000) -> Callable:
    """LFUCache: removes the least frequently used element.

    LFUCache：移除缓存中访问次数最少的元素，不论其访问时间。
    """
    return lambda func: cached(LFUCache(maxsize=maxsize))(func)


def fifo_cache(maxsize: int = 1000) -> Callable:
    """FIFOCache: removes the oldest element.

    FIFOCache：移除在缓存中停留时间最长的元素。
    """
    return lambda func: cached(FIFOCache(maxsize=maxsize))(func)


def rr_cache(maxsize: int = 1000) -> Callable:
    """RRCache: randomly removes an element when space is needed.

    RRCache: 在必要时随机选择一个元素进行移除，以腾出空间。
    """
    return lambda func: cached(RRCache(maxsize=maxsize))(func)



import hashlib
import os
import pickle
from functools import cached_property, wraps
from typing import Any, Callable, Optional

from funlog import getLogger

from ._utils import ensure_gitignore, normalize_args

logger = getLogger("funcache")

__all__ = ["PickleCache", "pkl_cache", "cached_property"]


class PickleCache:
    """Decorator that caches function results to pickle files on disk.

    Args:
        cache_key: The name of the function parameter whose value is used as the cache key.
        cache_dir: Directory to store pickle cache files.
        is_cache: The name of a boolean parameter that controls whether caching is enabled.
        printf: If True, also print cache log messages to stdout.
    """

    def __init__(
        self,
        cache_key: str,
        cache_dir: str = ".cache",
        is_cache: str = "cache",
        printf: bool = False,
    ) -> None:
        self.cache_key = cache_key
        self.cache_dir = cache_dir
        self.is_cache = is_cache
        self.printf = printf

    def _log(self, msg: str) -> None:
        if self.printf:
            print(msg)
        logger.debug(msg)

    def _get_cache_file(self, key: str) -> str:
        hashed = hashlib.md5(str(key).encode()).hexdigest()
        return os.path.join(self.cache_dir, f"{hashed}.pkl")

    @staticmethod
    def _load_cache(cache_file: str) -> Optional[Any]:
        try:
            with open(cache_file, "rb") as f:
                return pickle.load(f)
        except (FileNotFoundError, pickle.PickleError):
            return None

    def _save_cache(self, cache_file: str, data: Any) -> None:
        ensure_gitignore(self.cache_dir)
        with open(cache_file, "wb") as f:
            pickle.dump(data, f)

    def __call__(self, func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            merged = normalize_args(func, args, kwargs)

            is_cache = merged.get(self.is_cache, True)
            cache_key = merged.get(self.cache_key, "")
            is_cache = is_cache and cache_key is not None

            if is_cache:
                cache_file = self._get_cache_file(cache_key)
                cached_result = self._load_cache(cache_file)
                if cached_result is not None:
                    self._log(
                        f"Cache hit for function '{func.__name__}' with key: {cache_key}"
                    )
                    return cached_result

            result = func(**merged)

            if is_cache:
                self._save_cache(cache_file, result)
                self._log(
                    f"Cache data for function '{func.__name__}' with key: {cache_key}"
                )
            return result

        return wrapper


def pkl_cache(
    cache_key: str,
    cache_dir: str = ".cache",
    is_cache: str = "cache",
    printf: bool = False,
) -> PickleCache:
    """Convenience factory for :class:`PickleCache`."""
    return PickleCache(cache_key, cache_dir, is_cache, printf=printf)



import os
from functools import wraps
from hashlib import md5
from typing import Any, Callable, Optional

from diskcache import Cache
from funlog import getLogger

from ._utils import ensure_gitignore, normalize_args

logger = getLogger("funcache")

__all__ = ["DiskCache", "disk_cache"]


class DiskCache:
    """Decorator that caches function results using :class:`diskcache.Cache`.

    Args:
        cache_key: The name of the function parameter whose value is used as the cache key.
        cache_dir: Directory for the disk cache. Auto-generated from the function if *None*.
        is_cache: The name of a boolean parameter that controls whether caching is enabled.
        expire: Cache entry expiration time in seconds (default: 1 day).
    """

    def __init__(
        self,
        cache_key: str,
        cache_dir: Optional[str] = None,
        is_cache: str = "cache",
        expire: int = 60 * 60 * 24,
    ) -> None:
        self.cache_key = cache_key
        self.cache_dir = cache_dir
        self.is_cache = is_cache
        self.expire = expire
        self._cache: Optional[Cache] = None

    def _init_cache(self, func: Callable) -> Cache:
        if self._cache is not None:
            return self._cache

        if self.cache_dir is None:
            uid = md5(func.__code__.co_filename.encode("utf-8")).hexdigest()
            self.cache_dir = os.path.join(".disk_cache", f"{uid}-{func.__name__}")

        self._cache = Cache(self.cache_dir)
        ensure_gitignore(self.cache_dir)

        logger.success(
            f"init func {func.__name__} success. with cache_dir: {self.cache_dir}"
        )
        return self._cache

    def __call__(self, func: Callable) -> Callable:
        cache = self._init_cache(func)

        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            merged = normalize_args(func, args, kwargs)

            cache_key = merged.get(self.cache_key, "")
            is_cache = merged.get(self.is_cache, True) and cache_key is not None

            if not is_cache:
                return func(**merged)

            cached_result = cache.get(cache_key)
            if cached_result is not None:
                logger.debug(
                    f"Cache hit for function '{func.__name__}' with key: {cache_key}"
                )
                return cached_result

            result = func(**merged)
            cache.set(cache_key, result, expire=self.expire)
            logger.debug(
                f"Cache data for function '{func.__name__}' with key: {cache_key}"
            )
            return result

        return wrapper


def disk_cache(
    cache_key: str,
    cache_dir: Optional[str] = None,
    is_cache: str = "cache",
    expire: int = 60 * 60 * 24,
) -> DiskCache:
    """Convenience factory for :class:`DiskCache`."""
    return DiskCache(
        cache_key=cache_key,
        cache_dir=cache_dir,
        is_cache=is_cache,
        expire=expire,
    )



from .box import (
    cache,
    fifo_cache,
    lfu_cache,
    lru_cache,
    rr_cache,
    ttl_cache,
    vttl_cache,
)
from .core import PickleCache, cached_property, pkl_cache
from .disk import DiskCache, disk_cache

__all__ = [
    "cache",
    "lru_cache",
    "ttl_cache",
    "vttl_cache",
    "lfu_cache",
    "fifo_cache",
    "rr_cache",
    "PickleCache",
    "pkl_cache",
    "cached_property",
    "disk_cache",
    "DiskCache",
]
